import locale


def beautify_json(value):
    if type(value) == list:
        return [beautify_json(elem) for elem in value]
    elif type(value) == dict:
        keys = list(value.keys())
        keys.sort()
        return {key: beautify_json(value[key]) for key in keys}
    else:
        return value


def number_format(num, places=0):
    return locale.format_string("%.*f", (places, num), True)


def prepare_number(number):
    exp = 18
    numb = float(number) * pow(10, exp)
    prepared_num = number_format(numb, 0)
    return prepared_num



import base64
import json
from hashlib import sha256

import base58
import bech32

import datetime
import ethereum.transactions as crypto
import requests
import rlp
import sha3
from Cryptodome.Hash import SHA256
from ecdsa import SigningKey, SECP256k1
from ecdsa.util import sigencode_string_canonize

from decimal_sdk.transactions import RedeemCheckTransaction
from decimal_sdk.types import Coin
from .transactions import Transaction
from .types import FEES
from .utils.helpers import get_amount_uni, from_words
from .utils.fields_validator import validate_data
from .wallet import Wallet


class DecimalAPI:
    """
    Base class to perform operations on Decimal API.
    Create new instance of api with passing base URL to DecimalAPI class.
    """
    unit = 0.001

    def __init__(self, base_url: str):
        self.base_url = base_url
        if self.base_url[-1]:
            self.base_url += '/'

    def get_address(self, address: str):
        self.__validate_address(address)
        return self.__request(f'address/{address}')

    def get_coin(self, name: str):
        return self.__request(f'coin/{name}')

    def get_coins_list(self, limit: int = 10, offset: int = 0):
        options = {"limit": limit, "offset": offset}
        return self.__request('coin', 'get', None, options)

    def get_multisig(self, address: str):
        self.__validate_address(address)
        return self.__request(f'multisig/{address}')

    def get_multisigs(self, address: str, limit: int = 10, offset: int = 0):
        self.__validate_address(address)
        options = {"limit": limit, "offset": offset}
        return self.__request(f'address/{address}/multisigs', 'get', None, options)

    def get_my_transactions(self, wallet: Wallet, limit: int = 10, offset: int = 0):
        options = {"limit": limit, "offset": offset}
        return self.__request(f'address/{wallet.get_address()}/txs', 'get', None, options)

    def get_nonce(self, address: str):
        self.__validate_address(address)
        return self.__request(f'rpc/auth/accounts/{address}')

    def get_nonce_not_increasing(self, address: str):
        self.__validate_address(address)
        return self.__request(f'rpc/accounts/{address}')

    def get_stakes(self, address: str):
        self.__validate_address(address)
        return self.__request(f'address/{address}/stakes')

    def get_tx(self, tx_hash: str):
        return json.loads(self.__request("rpc/tx", "get", options={"hash": f"{tx_hash}"}))["result"]

    def get_txs_multisign(self, address: str, limit: int = 10, offset: int = 0):
        options = {"limit": limit, "offset": offset}
        self.__validate_address(address)
        return self.__request(f"multisig/{address}/txs", 'get', None, options)

    def get_validator(self, address: str):
        self.__validate_address(address)
        return self.__request(f'validator/{address}')

    def get_nft(self, id: str, wallet: Wallet):
        timestamp = int(str(datetime.datetime.now().timestamp()).split(".")[0])
        data = {
            "nftId": id,
            "timestamp": timestamp
        }

        encoded = json.dumps(data).replace(" ", "").encode()
        msg_hash = sha3.keccak_256(encoded)
        v, r, s = crypto.ecsign(msg_hash.digest(), wallet.get_private_key())

        rr = int_to_big_endian(r).hex()
        ss = int_to_big_endian(s).hex()
        vv = v-27

        options={
            "timestamp": timestamp,
            "signature": json.dumps({
                "r":rr,
                "s":ss,
                "recoveryParam":vv})
        }
        return self.__request(f'nfts/{id}', 'get', None, options)

    def get_nft_stakes(self, address: str):
        self.__validate_address(address)
        return self.__request(f'address/{address}/nfts/stakes')

    def estimate_tx_fee(self, tx: Transaction, wallet: Wallet, options={}):
        """Method to sign and send prepared transaction"""
        url = "rpc/txs"

        denom = "del"
        commission_type = "base"
        if "denom" in options:
            denom = options["denom"]
            commission_type = "value"

        if "memo" in options:
            tx.memo = options["memo"]
            tx.signer.memo = options["memo"]

        message = tx.memo
        tx_data = tx.message.get_message()
        return self.__get_comission(tx, denom, FEES[tx.message.type], tx_data)["value"]

    def send_tx(self, tx: Transaction, wallet: Wallet, options={}):
        """Method to sign and send prepared transaction"""
        url = "rpc/txs"

        denom = "del"
        commission_type = "base"
        if "denom" in options:
            denom = options["denom"]
            commission_type = "value"

        if "memo" in options:
            options["memo"] = options["memo"]
            tx.memo = options["memo"]
            tx.signer.memo = options["memo"]

        message = tx.memo
        tx_data = tx.message.get_message()
        validate_data(tx_data["value"])
        commission = self.__get_comission(tx, denom, FEES[tx.message.type], tx_data)
        if commission["fee_in_custom"] == 0:
            raise Exception("Coin for custom fee not found")

        fee_amount = Coin(denom.lower(), get_amount_uni(commission[commission_type]))
        wallet.nonce = json.loads(self.get_nonce_not_increasing(wallet.get_address()))["result"]
        tx.signer.chain_id = self.get_chain_id()
        tx.signer.account_number = str(wallet.nonce["value"]["account_number"])
        tx.signer.sequence = str(wallet.nonce["value"]["sequence"])

        if denom == "del" or denom == "tdel":
            tx.fee.amount = []
            tx.signer.fee.amount = []
        else:
            tx.fee.amount = [fee_amount]
            tx.signer.fee.amount = [fee_amount]

        tx.signer.fee = tx.fee
        tx.signer.msgs = tx.msgs
        tx.signer.memo = tx.memo

        payload = {"tx": {}, "mode": "sync"}
        payload["tx"]["msg"] = [tx_data]
        # payload["tx"]["fee"] = {"amount": [{'denom': 'del', 'amount': '0'}], "gas": "0"}
        if denom == "del" or denom == "tdel":
            payload["tx"]["fee"] = {"amount": [], "gas": "0"}
        else:
            payload["tx"]["fee"] = {"amount": [tx.signer.fee.amount[0].__dict__()], "gas": "0"}
        payload["tx"]["memo"] = message
        payload["tx"]["signatures"] = []
        tx.sign(wallet)
        tx.msgs.clear()
        tx.signer.msgs.clear()

        for sig in tx.signatures:
            payload["tx"]["signatures"].append(sig.get_signature())

        return self.__request(url, 'post', json.dumps(payload))

    def issue_check(self, wallet, data):
        new_data = {
            "coin": data["coin"].lower(),
            "amount": int(get_amount_uni(int(data["amount"]), False)),
            "nonce": data["nonce"],
            "due_block": int(data["due_block"]),
            "passphrase": data["password"],
        }

        chain_id = self.get_chain_id()

        passphrase_hash = SHA256.new(str.encode(new_data["passphrase"])).digest()
        pp = []
        for b in passphrase_hash:
            pp.append(b)

        check_hash = self.__rpl_hash([
            chain_id,
            new_data["coin"],
            new_data["amount"],
            new_data["nonce"],
            new_data["due_block"],
        ])

        ch = []
        for b in check_hash:
            ch.append(b)

        passphrase_hash = sha256(str.encode(new_data["passphrase"])).digest()

        sk = SigningKey.from_string(passphrase_hash, curve=SECP256k1)
        lock_obj = sk.sign_digest_deterministic(check_hash, hashfunc=sha256, sigencode=sigencode_string_canonize)

        bts2 = []
        for b in lock_obj:
            bts2.append(b)
        lock_signature = bytearray(65)

        i = 0
        while i < 64:
            lock_signature[i] = lock_obj[i]
            i += 1

        v, r, s = crypto.ecsign(check_hash, passphrase_hash)
        lock_signature[64] = v - 27

        check_locked_hash = self.__rpl_hash([
            chain_id,
            new_data["coin"],
            new_data["amount"],
            new_data["nonce"],
            new_data["due_block"],
            lock_signature
        ])
        clh = []
        for b in check_locked_hash:
            clh.append(b)
        pk = []
        for b in wallet.get_private_key():
            pk.append(b)
        sk = SigningKey.from_string(wallet.get_private_key(), curve=SECP256k1)
        check_obj = sk.sign_digest_deterministic(check_locked_hash, hashfunc=sha256,
                                                 sigencode=sigencode_string_canonize)
        co = []
        for b in check_obj:
            co.append(b)
        v, r, s = crypto.ecsign(check_locked_hash, wallet.get_private_key())

        rr = bytearray(32)
        ss = bytearray(32)

        i = 0
        while i < 32:
            rr[i] = crypto.int_to_32bytearray(r)[i]
            i += 1

        i = 0
        while i < 32:
            ss[i] = crypto.int_to_32bytearray(s)[i]
            i += 1

        check = rlp.encode([
            chain_id,
            new_data["coin"],
            new_data["amount"],
            new_data["nonce"],
            new_data["due_block"],
            lock_signature,
            v,
            rr,
            ss,
        ])
        chk = []
        for b in check:
            chk.append(b)
        return base58.b58encode(check).decode('utf-8')

    def redeem_check(self, data, wallet):
        passphrase_hash = SHA256.new(str.encode(data["password"])).digest()

        words = bech32.bech32_decode(wallet.get_address())
        f_words = from_words(words[1])

        f_w_buffer = bytearray()
        for b in f_words:
            f_w_buffer.append(b)

        sender_address_hash = self.__rpl_hash([f_w_buffer])
        sah = []
        for b in sender_address_hash:
            sah.append(b)

        sk = SigningKey.from_string(passphrase_hash, curve=SECP256k1)
        proof_obj = sk.sign_digest_deterministic(sender_address_hash, hashfunc=sha256,
                                                 sigencode=sigencode_string_canonize)

        po = []
        for b in proof_obj:
            po.append(b)

        proof_signature = bytearray(65)

        i = 0
        while i < 64:
            proof_signature[i] = proof_obj[i]
            i += 1

        v, r, s = crypto.ecsign(sender_address_hash, passphrase_hash)
        proof_signature[64] = v - 27

        proof = proof_signature
        proof = base64.b64encode(proof)

        transaction = RedeemCheckTransaction(wallet.get_address(), data["check"], proof.decode('utf-8'))

        return self.send_tx(transaction, wallet)

    def get_chain_id(self):
        url = "rpc/node_info"
        resp = json.loads(self.__request(url))
        chain_id = resp["node_info"]["network"]
        return chain_id

    @staticmethod
    def __validate_address(address: str):
        if len(address) < 41 or not address.startswith('dx'):
            raise Exception('Invalid address')

    @staticmethod
    def __rpl_hash(data):
        khash = sha3.keccak_256()
        khash.update(rlp.encode(data))
        return khash.digest()

    def get_coin_price(self, name: str):
        coin = json.loads(self.get_coin(name))
        if not coin["ok"]:
            raise Exception('Coin not found')
        coin = coin["result"]
        reserve = int(get_amount_uni(int(coin["reserve"]), True))
        supply = int(get_amount_uni(int(coin["volume"]), True))
        crr = int(coin["crr"]) / 100

        amount = min(supply, 1)

        if supply == 0:
            return 0

        result = amount / supply
        result = 1 - result
        result = pow(result, 1 / crr)
        result = (1 - result) * reserve * 1.03

        return result

    def __get_tx_size(self, tx: Transaction):
        value = {"msg": [tx.message.get_message()]}
        value["fee"] = {"amount": [], "gas": "0"}
        value["memo"] = tx.memo
        preparedTx = {
            "type": 'cosmos-sdk/StdTx',
            "value": value
        }

        signatureSize = 109
        resp = json.loads(self.__request('rpc/txs/encode', 'post', json.dumps(preparedTx)))

        encoded_tx_base64 = resp["tx"]
        encoded_tx = len(base64.b64decode(encoded_tx_base64))
        size = encoded_tx + signatureSize
        return size

    def __get_comission(self, tx: Transaction, fee_coin, operation_fee, tx_data):
        if tx.message.get_type() == 'coin/create_coin':
            ticker_length = len(tx_data['value']['symbol'])
            if ticker_length == 3:
                operation_fee = 1000000
            elif ticker_length == 4:
                operation_fee = 100000
            elif ticker_length == 5:
                operation_fee = 10000
            elif ticker_length == 6:
                operation_fee = 1000
            else:
                operation_fee = 100

        ticker = fee_coin
        text_size = self.__get_tx_size(tx)
        fee_for_text = text_size * 2
        fee_in_base = (operation_fee + fee_for_text + 20) / 1000

        if tx.message.get_type() == 'coin/multi_send_coin':
            number_of_participants = len(tx.message.get_message()["value"]["sends"])
            fee_for_participants = 5 * (number_of_participants - 1) / 1000
            fee_in_base = fee_in_base + fee_for_participants

        if fee_coin in ['del', 'tdel']:
            return {"coinPrice": "1", "value": fee_in_base, "base": fee_in_base}

        try:
            coin_price = self.get_coin_price(ticker)
            fee_in_custom = fee_in_base / coin_price

        except:
            fee_in_custom = 0
            coin_price = 0

        return {"coinPrice": str(coin_price), 'value': fee_in_custom, 'base': fee_in_base}

    def __request(self, path: str, method: str = 'get', payload=None, options={}):
        url = (self.base_url + path).lower()
        if method == 'get':
            if len(options) > 0:
                response = requests.get(url, params=options)
            else:
                response = requests.get(url)
        else:
            response = requests.post(url, payload)
        return response.text

def int_to_big_endian(value: int) -> bytes:
    return value.to_bytes((value.bit_length() + 7) // 8 or 1, "big")



import time
from decimal_sdk.msgs.base import BaseMsg
from decimal_sdk.tx_types import *
from decimal_sdk.types import Coin, Candidate, Signature
from ..utils.helpers import get_amount_uni


class SendCoinMsg(BaseMsg):
    type = COIN_SEND
    sender: str
    receiver: str
    coin: Coin

    def __init__(self, sender: str, receiver: str, coin: Coin):
        self.sender = sender
        self.receiver = receiver
        self.coin = coin

    def __dict__(self):
        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'receiver': self.receiver,
                'coin': self.coin.__dict__()
            }
        }


class BuyCoinMsg(BaseMsg):
    type = COIN_BUY
    sender: str
    coin_to_buy: Coin
    max_coin_to_sell: Coin

    def __init__(self, sender: str, coin_to_buy: Coin, max_coin_to_sell: Coin):
        self.sender = sender
        self.coin_to_buy = coin_to_buy
        self.max_coin_to_sell = max_coin_to_sell

    def __dict__(self):
        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'coin_to_buy': self.coin_to_buy.__dict__(),
                'max_coin_to_sell': self.max_coin_to_sell.__dict__()
            }
        }


class SellCoinMsg(BaseMsg):
    type = COIN_SELL
    sender: str
    coin_to_sell: Coin
    min_coin_to_buy: Coin

    def __init__(self, sender: str, coin_to_sell: Coin, min_coin_to_buy: Coin):
        self.sender = sender
        self.coin_to_sell = coin_to_sell
        self.min_coin_to_buy = min_coin_to_buy

    def __dict__(self):
        return {'type': self.type,
                'value': {
                    'sender': self.sender,
                    'coin_to_sell': self.coin_to_sell.__dict__(),
                    'min_coin_to_buy': self.min_coin_to_buy.__dict__()
                }}


class CreateCoinMsg(BaseMsg):
    type = COIN_CREATE
    sender: str
    title: str
    symbol: str
    crr: int
    initial_volume: str
    initial_reserve: str
    identity: str
    limit_volume: str

    def __init__(self,
                 sender: str,
                 title: str,
                 symbol: str,
                 crr: int,
                 initial_volume: str,
                 initial_reserve: str,
                 identity: str,
                 limit_volume: str):
        self.sender = sender
        self.title = title
        self.symbol = symbol
        self.crr = crr
        self.initial_volume = initial_volume
        self.initial_reserve = initial_reserve
        self.identity = identity
        self.limit_volume = limit_volume

    def __dict__(self):
        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'title': self.title,
                'symbol': self.symbol,
                'constant_reserve_ratio': self.crr,
                'initial_volume': self.initial_volume,
                'initial_reserve': self.initial_reserve,
                'identity': self.identity,
                'limit_volume': self.limit_volume
                }
            }


class UpdateCoinMsg(BaseMsg):
    type = COIN_UPDATE
    sender: str
    symbol: str
    identity: str
    limit_volume: str

    def __init__(self,
                 sender: str,
                 symbol: str,
                 identity: str,
                 limit_volume: str):
        self.sender = sender
        self.symbol = symbol
        self.identity = identity
        self.limit_volume = limit_volume

    def __dict__(self):
        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'symbol': self.symbol,
                'identity': self.identity,
                'limit_volume': self.limit_volume
                }
            }


class SellAllCoinsMsg(BaseMsg):
    type = COIN_SELL_ALL
    sender: str
    coin_to_sell: Coin
    min_coin_to_buy: Coin

    def __init__(self, sender: str, coin_to_sell: Coin, min_coin_to_buy: Coin):
        self.sender = sender
        self.coin_to_sell = coin_to_sell
        self.min_coin_to_buy = min_coin_to_buy

    def __dict__(self):
        return {'type': self.type,
                'value': {
                    'sender': self.sender,
                    'coin_to_sell': self.coin_to_sell.__dict__(),
                    'min_coin_to_buy': self.min_coin_to_buy.__dict__()
                }}


class DelegateMsg(BaseMsg):
    type = VALIDATOR_DELEGATE
    delegator_address: str
    validator_address: str
    coin: Coin

    def __init__(self, delegator_address: str, validator_address: str, coin: Coin):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.coin = coin

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "coin": self.coin.__dict__()
            }
        }


class UnboundMsg(BaseMsg):
    type = VALIDATOR_UNBOND
    delegator_address: str
    validator_address: str
    coin: Coin

    def __init__(self, delegator_address: str, validator_address: str, coin: Coin):
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.coin = coin

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "coin": self.coin.__dict__()
            }
        }


class DeclareCandidateMsg(BaseMsg):
    type = VALIDATOR_CANDIDATE
    commission: int
    validator_addr: str
    reward_addr: str
    pub_key: Signature
    stake: Coin
    description: Candidate

    def __init__(self, commission: int, validator_addr: str, reward_addr: str, pub_key: Signature,
                 stake: Coin, description: Candidate):
        self.commission = commission
        self.validator_addr = validator_addr
        self.reward_addr = reward_addr
        self.pub_key = pub_key
        self.stake = stake
        self.description = description

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "commission": str(self.commission),
                "validator_addr": self.validator_addr,
                "reward_addr": self.reward_addr,
                "pub_key": self.pub_key,
                "stake": self.stake.__dict__(),
                "description": self.description.__dict__()
            }
        }


class EditCandidateMsg(BaseMsg):
    type = VALIDATOR_CANDIDATE_EDIT
    validator_address: str
    reward_address: str
    description: Candidate

    def __init__(self, validator_address: str, reward_address: str, description: Candidate):
        self.validator_address = validator_address
        self.reward_address = reward_address
        self.description = description

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "validator_address": self.validator_address,
                "reward_address": self.reward_address,
                "description": self.description.__dict__()
            }
        }


class DisableEnableValidatorMsg(BaseMsg):
    type: str
    validator_address: str

    state = {
        "disable": VALIDATOR_SET_OFFLINE,
        "enable": VALIDATOR_SET_ONLINE
    }

    def __init__(self, set_state: str, validator_address: str):
        self.type = self.state[set_state]
        self.validator_address = validator_address

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "validator_address": self.validator_address
            }
        }


class MultysigCreateMsg(BaseMsg):
    type = MULTISIG_CREATE_WALLET
    sender: str
    owners: list
    weights: list
    threshold: int

    def __init__(self, sender: str, owners: list, weights: list, threshold: int):
        self.sender = sender
        self.owners = owners
        self.weights = weights
        self.threshold = threshold

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "sender": self.sender,
                "owners": self.owners,
                "weights": self.weights,
                "threshold": self.threshold
            }
        }


class MultysigCreateTXMsg(BaseMsg):
    type = MULTISIG_CREATE_TX
    sender: str
    wallet: str
    receiver: str
    coins: Coin

    def __init__(self, sender: str, wallet: str, receiver: str, coins: Coin):
        self.sender = sender
        self.wallet = wallet
        self.receiver = receiver
        self.coins = coins

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "sender": self.sender,
                "wallet": self.wallet,
                "receiver": self.receiver,
                "coins": self.coins.__dict__()
            }
        }


class MultysigSignTXMsg(BaseMsg):
    type = MULTISIG_SIGN_TX

    sender: str
    tx_id: str

    def __init__(self, sender: str, tx_id: str):
        self.sender = sender
        self.tx_id = tx_id

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "sender": self.sender,
                "tx_id": self.tx_id
            }
        }


class MultisendSend:
    receiver: str
    coin: Coin

    def __init__(self, receiver: str, coin: Coin):
        self.receiver = receiver
        self.coin = coin

    def __dict__(self):
        return {
            "coin": self.coin.__dict__(),
            "receiver": self.receiver
        }


class MultisendCoinMsg(BaseMsg):
    type = COIN_MULTISEND

    sender: str
    sends: [MultisendSend]

    def __init__(self, sender: str, sends: [MultisendSend]):
        self.sender = sender
        self.sends = sends

    def __dict__(self):
        snds = []
        for send in self.sends:
            snds.append({
                "receiver": send["receiver"],
                "coin": {
                    "amount": get_amount_uni(send["amount"]),
                    "denom": send["coin"]
                }
            })

        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'sends': snds
            }
        }


class SubmitProposalMsg(BaseMsg):
    type = PROPOSAL_SUBMIT
    content: str
    proposer: str
    voting_start_block: str
    voting_end_block: str

    def __init__(self, content: str, proposer: str, voting_start_block: str, voting_end_block: str):
        self.content = content
        self.proposer = proposer
        self.voting_start_block = voting_start_block
        self.voting_end_block = voting_end_block

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "content": self.content,
                "proposer": self.proposer,
                "voting_start_block": self.voting_start_block,
                "voting_end_block": self.voting_end_block,
            }
        }


class VoteProposalMsg(BaseMsg):
    type = PROPOSAL_VOTE
    proposal_id: int
    voter: str
    option: str

    def __init__(self, proposal_id: int, voter: str, option: str):
        self.proposal_id = proposal_id
        self.voter = voter
        self.option = option

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "proposal_id": self.proposal_id,
                "voter": self.voter,
                "option": self.option
            }
        }


class RedeemCheckMsg(BaseMsg):
    type = COIN_REDEEM_CHECK
    sender: str
    check: str
    proof: str

    def __init__(self, sender: str, check: str, proof: Coin):
        self.sender = sender
        self.check = check
        self.proof = proof

    def __dict__(self):
        return {
            'type': self.type,
            'value': {
                'sender': self.sender,
                'check': self.check,
                'proof': self.proof
            }
        }


class SwapRedeemMsg(BaseMsg):
    type = SWAP_REDEEM
    sender: str
    sent_from: str
    recipient: str
    amount: str
    token_name: str
    token_symbol: str
    from_chain: str
    dest_chain: str
    v: str
    r: str
    s: str

    def __init__(self, sender: str,
                 sent_from: str,
                 recipient: str,
                 amount: str,
                 token_name: str,
                 token_symbol: str,
                 from_chain: str,
                 dest_chain: str,
                 v: str, r: str, s: str):
        self.sender = sender
        self.sent_from = sent_from
        self.recipient = recipient
        self.amount = amount
        self.token_name = token_name
        self.token_symbol = token_symbol
        self.from_chain = from_chain
        self.dest_chain = dest_chain
        self.v = v
        self.r = r
        self.s = s

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                'sender': self.sender,
                'from': self.sender,
                'recipient': self.recipient,
                'amount': self.amount,
                'token_name': self.token_name,
                'token_symbol': self.token_symbol,
                'transaction_number': str(round(time.time() * 1000)),
                'from_chain': self.from_chain,
                'dest_chain': self.dest_chain,
                'v': self.v,
                'r': self.r,
                's': self.s,
            }
        }


class SwapInitMsg(BaseMsg):
    type = SWAP_INIT
    sender: str
    recipient: str
    amount: str
    token_symbol: str
    from_chain: str
    dest_chain: str

    def __init__(self, sender: str, recipient: str, amount: str, token_symbol: str, from_chain: str, dest_chain: str):
        self.sender = sender
        self.recipient = recipient
        self.amount = amount
        self.token_symbol = token_symbol
        self.from_chain = from_chain
        self.dest_chain = dest_chain

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                'from': self.sender,
                'recipient': self.recipient,
                'amount': self.amount,
                'token_symbol': self.token_symbol,
                'transaction_number': str(round(time.time() * 1000)),
                'from_chain': self.dest_chain,
                'dest_chain': self.dest_chain,
            }
        }


class NftMintMsg(BaseMsg):
    type = NFT_MINT
    denom: str
    id: str
    sender: str
    recipient: str
    quantity: str
    reserve: str
    token_uri: str
    allow_mint: bool

    def __init__(self, denom: str, id: str, sender: str, recipient: str, quantity: int, reserve: int, token_uri: str,
                 allow_mint: bool):
        if not isinstance(quantity, int):
            raise ValueError('Quantity must be of type int')
        self.denom = denom
        self.id = id
        self.sender = sender
        self.recipient = recipient
        self.quantity = str(quantity)
        self.reserve = str(get_amount_uni(reserve))
        self.token_uri = token_uri
        self.allow_mint = allow_mint

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "sender": self.sender,
                "recipient": self.recipient,
                "quantity": self.quantity,
                "reserve": self.reserve,
                "token_uri": self.token_uri,
                "allow_mint": self.allow_mint,
            }
        }


class NftBurnMsg(BaseMsg):
    type = NFT_BURN
    denom: str
    id: str
    sender: str
    sub_token_ids: []

    def __init__(self, denom: str, id: str, sender: str, sub_token_ids: []):
        self.denom = denom
        self.id = id
        self.sender = sender
        self.sub_token_ids = sub_token_ids

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "sender": self.sender,
                "sub_token_ids": self.sub_token_ids,
            }
        }


class NftEditMetadataMsg(BaseMsg):
    type = NFT_EDIT_METADATA
    denom: str
    id: str
    sender: str
    token_uri: str

    def __init__(self, denom: str, id: str, sender: str, token_uri: str):
        self.denom = denom
        self.id = id
        self.sender = sender
        self.token_uri = token_uri

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "sender": self.sender,
                "token_uri": self.token_uri,
            }
        }


class NftTransferMsg(BaseMsg):
    type = NFT_TRANSFER
    denom: str
    id: str
    sender: str
    recipient: str
    sub_token_ids: []

    def __init__(self, denom: str, id: str, sender: str, recipient: str, sub_token_ids: []):
        self.denom = denom
        self.id = id
        self.sender = sender
        self.recipient = recipient
        self.sub_token_ids = sub_token_ids

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "sender": self.sender,
                "recipient": self.recipient,
                "sub_token_ids": self.sub_token_ids,
            }
        }


class NftDelegateMsg(BaseMsg):
    type = NFT_DELEGATE
    denom: str
    id: str
    delegator_address: str
    validator_address: str
    sub_token_ids: []

    def __init__(self, denom: str, id: str, delegator_address: str, validator_address: str, sub_token_ids: []):
        self.denom = denom
        self.id = id
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.sub_token_ids = sub_token_ids

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "sub_token_ids": self.sub_token_ids,
            }
        }


class NftUnboundMsg(BaseMsg):
    type = NFT_UNBOND
    denom: str
    id: str
    delegator_address: str
    validator_address: str
    sub_token_ids: []

    def __init__(self, denom: str, id: str, delegator_address: str, validator_address: str, sub_token_ids: []):
        self.denom = denom
        self.id = id
        self.delegator_address = delegator_address
        self.validator_address = validator_address
        self.sub_token_ids = sub_token_ids

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "delegator_address": self.delegator_address,
                "validator_address": self.validator_address,
                "sub_token_ids": self.sub_token_ids,
            }
        }


class NftUpdateReserveMsg(BaseMsg):
    type = NFT_UPDATE_RESERVE
    sender: str
    id: str
    denom: str
    sub_token_ids: []
    reserve: str

    def __init__(self, denom: str, id: str, sender: str, reserve: str, sub_token_ids: []):
        self.denom = denom
        self.id = id
        self.sender = sender
        self.sub_token_ids = sub_token_ids
        self.reserve = str(get_amount_uni(reserve))

    def __dict__(self):
        return {
            "type": self.type,
            "value": {
                "denom": self.denom,
                "id": self.id,
                "sender": self.sender,
                "reserve": self.reserve,
                "sub_token_ids": self.sub_token_ids,
            }
        }



from .wallet import Wallet
from decimal_sdk.msgs.base import BaseMsg
from decimal_sdk.types import StdSignMsg, SignMeta, Fee, Coin, Candidate
from decimal_sdk.utils.helpers import get_amount_uni
from decimal_sdk.msgs.msgs import (SendCoinMsg, BuyCoinMsg, CreateCoinMsg, UpdateCoinMsg, SellAllCoinsMsg,
                                   DelegateMsg, UnboundMsg, RedeemCheckMsg,
                                   DeclareCandidateMsg, EditCandidateMsg,
                                   DisableEnableValidatorMsg,
                                   MultysigCreateMsg, MultysigCreateTXMsg, MultysigSignTXMsg, MultisendSend, MultisendCoinMsg,
                                   SubmitProposalMsg, VoteProposalMsg,
                                   SwapRedeemMsg, SwapInitMsg,
                                   NftMintMsg, NftBurnMsg, NftEditMetadataMsg, NftTransferMsg, NftDelegateMsg, NftUnboundMsg, NftUpdateReserveMsg,
                                   )


class Transaction:
    fee: Fee
    memo: str
    meta: SignMeta
    signatures = []
    msgs = []
    signer: StdSignMsg
    message: BaseMsg

    def __init__(self, **kwargs):
        self.meta = SignMeta()
        self.fee = Fee([], '0')
        self.memo = ''
        self.signer = StdSignMsg(self, meta=self.meta)
        self.signer.add_msg(self.message)

    def add_msg(self, msg: BaseMsg):
        self.msgs.append(msg)

    def sign(self, wallet: Wallet):
        self.signer.sign(wallet)
        self.signatures = self.signer.signatures


class SendCoinTransaction(Transaction):
    message: SendCoinMsg

    def __init__(self, sender, receiver, denom, amount, **kwargs):
        coin = Coin(denom, get_amount_uni(amount))
        self.message = SendCoinMsg(sender, receiver, coin)
        super().__init__(**kwargs)


class BuyCoinTransaction(Transaction):
    message: BuyCoinMsg

    def __init__(self, sender, coin_to_buy, coin_to_spend, amount_to_buy, limit=100000000000, **kwargs):
        coin_to_buy = Coin(coin_to_buy, get_amount_uni(amount_to_buy))
        max_coin_to_sell = Coin(coin_to_spend, get_amount_uni(limit))
        self.message = BuyCoinMsg(sender, coin_to_buy, max_coin_to_sell)
        super().__init__(**kwargs)


class CreateCoinTransaction(Transaction):
    message: CreateCoinMsg

    def __init__(self, sender, title, symbol, crr, initial_volume, initial_reserve, identity, limit_volume, **kwargs):
        self.message = CreateCoinMsg(sender, title, symbol, crr, initial_volume, initial_reserve, identity, limit_volume)
        super().__init__(**kwargs)


class UpdateCoinTransaction(Transaction):
    message: UpdateCoinMsg

    def __init__(self, sender, symbol, identity, limit_volume, **kwargs):
        self.message = UpdateCoinMsg(sender, symbol, identity, limit_volume)
        super().__init__(**kwargs)


class SellAllCoinsMsgTransaction(Transaction):
    message: SellAllCoinsMsg

    def __init__(self, sender: str, coin_to_sell_denom: str, coin_to_sell_amount: int, min_coin_to_buy_denom: str,
                 min_coin_to_buy_amount=0, **kwargs):
        coin_to_sell_amount = str(get_amount_uni(coin_to_sell_amount))
        coin_to_sell = Coin(coin_to_sell_denom, coin_to_sell_amount)
        if min_coin_to_buy_amount == 0:
            min_coin_to_buy_amount = 1 #str(get_amount_uni(1))
        else:
            min_coin_to_buy_amount = str(get_amount_uni(min_coin_to_buy_amount))
        min_coin_to_buy = Coin(min_coin_to_buy_denom, min_coin_to_buy_amount)
        self.message = SellAllCoinsMsg(sender, coin_to_sell, min_coin_to_buy)
        super().__init__(**kwargs)


class DelegateTransaction(Transaction):
    message: DelegateMsg

    def __init__(self, delegator_address: str, validator_address: str, denom: str, amount: str, **kwargs):
        coin = Coin(denom, amount)
        self.message = DelegateMsg(delegator_address, validator_address, coin)
        super().__init__(**kwargs)


class UnbondTransaction(Transaction):
    message: UnboundMsg

    def __init__(self, delegator_address: str, validator_address: str, denom: str, amount: str, **kwargs):
        coin = Coin(denom, amount)
        self.message = UnboundMsg(delegator_address, validator_address, coin)
        super().__init__(**kwargs)


class RedeemCheckTransaction(Transaction):
    message: RedeemCheckMsg

    def __init__(self, sender, check, proof, **kwargs):
        self.message = RedeemCheckMsg(sender, check, proof)
        super().__init__(**kwargs)


class DeclareCandidateTransaction(Transaction):
    message: DeclareCandidateMsg

    def __init__(self, commission: int, validator_addr: str, reward_addr: str,
                 denom: str, amount: int,
                 moniker: str, identity: str, website: str, security_contact: str, details: str,
                 key_value: str, key_type: str = 'tendermint/PubKeyEd25519', **kwargs):
        stake = Coin(denom, str(amount))
        pub_key = {"type": key_type, "value": key_value}
        commission = '%.18f' % int(commission)
        description = Candidate(moniker, identity, website, security_contact, details)
        self.message = DeclareCandidateMsg(commission, validator_addr, reward_addr, pub_key, stake, description)
        super().__init__(**kwargs)


class EditCandidateTransaction(Transaction):
    message: EditCandidateMsg

    def __init__(self, validator_address: str, reward_address: str,
                 moniker: str, identity: str, website: str, security_contact: str, details: str, **kwargs):
        description = Candidate(moniker, identity, website, security_contact, details)
        self.message = EditCandidateMsg(validator_address, reward_address, description)
        super().__init__(**kwargs)


class DisableEnableValidatorTransaction(Transaction):
    message: DisableEnableValidatorMsg

    def __init__(self, set_state: str, validator_address: str, **kwargs):
        self.message = DisableEnableValidatorMsg(set_state, validator_address)
        super().__init__(**kwargs)


class MultysigCreateTransaction(Transaction):
    message: MultysigCreateMsg

    def __init__(self, sender: str, owners: list, weights: list, threshold: int, **kwargs):
        self.message = MultysigCreateMsg(sender, owners, weights, threshold)
        super().__init__(**kwargs)


class MultysigCreateTXTransaction(Transaction):
    message: MultysigCreateTXMsg

    def __init__(self, sender: str, wallet: str, receiver: str, denom: str, amount: str, **kwargs):
        coin = Coin(denom, amount)
        self.message = MultysigCreateTXMsg(sender, wallet, receiver, coin)
        super().__init__(**kwargs)


class MultysigSignTXTransaction(Transaction):
    message: MultysigSignTXMsg

    def __init__(self, sender: str, tx_id: str, **kwargs):
        self.message = MultysigSignTXMsg(sender, tx_id)
        super().__init__(**kwargs)


class MultisendCoinTransaction(Transaction):
    message: MultisendCoinMsg

    def __init__(self, sender: str, sends: [MultisendSend], **kwargs):
        self.message = MultisendCoinMsg(sender, sends)
        super().__init__(**kwargs)


class SubmitProposalTransaction(Transaction):
    message: SubmitProposalMsg

    def __init__(self, content: str, proposer: str, voting_start_block: str, voting_end_block: str, **kwargs):
        self.message = SubmitProposalMsg(content, proposer, voting_start_block, voting_end_block)
        super().__init__(**kwargs)


class VoteProposalTransaction(Transaction):
    message: VoteProposalMsg

    def __init__(self, proposal_id: int, voter: str, option: str, **kwargs):
        self.message = VoteProposalMsg(proposal_id, voter, option)
        super().__init__(**kwargs)


class NftMintTransaction(Transaction):
    message: NftMintMsg

    def __init__(self, denom: str, id: str, sender: str, recipient: str, quantity: int, reserve: int, token_uri: str,
                 allow_mint: bool, **kwargs):
        self.message = NftMintMsg(denom, id, sender, recipient, quantity, reserve, token_uri, allow_mint)
        super().__init__(**kwargs)


class NftBurnTransaction(Transaction):
    message: NftBurnMsg

    def __init__(self, denom: str, id: str, sender: str, sub_token_ids: [], **kwargs):
        self.message = NftBurnMsg(denom, id, sender, sub_token_ids)
        super().__init__(**kwargs)


class NftEditMetadataTransaction(Transaction):
    message: NftEditMetadataMsg

    def __init__(self, denom: str, id: str, sender: str, token_uri: str, **kwargs):
        self.message = NftEditMetadataMsg(denom, id, sender, token_uri)
        super().__init__(**kwargs)


class NftTransferTransaction(Transaction):
    message: NftTransferMsg

    def __init__(self, denom: str, id: str, sender: str, recipient: str, sub_token_ids: [], **kwargs):
        self.message = NftTransferMsg(denom, id, sender, recipient, sub_token_ids)
        super().__init__(**kwargs)


class NftDelegateTransaction(Transaction):
    message: NftDelegateMsg

    def __init__(self, denom: str, id: str, delegator_address: str, validator_address: str, sub_token_ids: [], **kwargs):
        self.message = NftDelegateMsg(denom, id, delegator_address, validator_address, sub_token_ids)
        super().__init__(**kwargs)


class NftUnboundTransaction(Transaction):
    message: NftUnboundMsg

    def __init__(self, denom: str, id: str, delegator_address: str, validator_address: str, sub_token_ids: [], **kwargs):
        self.message = NftUnboundMsg(denom, id, delegator_address, validator_address, sub_token_ids)
        super().__init__(**kwargs)


class NftUpdateReserveTransaction(Transaction):
    message: NftUpdateReserveMsg

    def __init__(self, denom: str, id: str, sender: str, reserve: str, sub_token_ids: [], **kwargs):
        self.message = NftUpdateReserveMsg(denom, id, sender, reserve, sub_token_ids)
        super().__init__(**kwargs)


class SwapRedeemTransaction(Transaction):
    message: SwapRedeemMsg

    def __init__(self, sender: str, sent_from: str, recipient: str, amount: int, token_name: str, token_symbol: str,
                 from_chain: str, dest_chain: str, v: str, r: str, s: str, **kwargs):
        self.message = SwapRedeemMsg(sender, sent_from, recipient, str(amount), token_name, token_symbol, from_chain, dest_chain, v, r, s)
        super().__init__(**kwargs)


class SwapInitTransaction(Transaction):
    message: SwapInitMsg

    def __init__(self, sender: str, recipient: str, amount: int, token_symbol: str, from_chain: str, dest_chain, **kwargs):
        self.message = SwapInitMsg(sender, recipient, str(amount), token_symbol, from_chain, dest_chain)
        super().__init__(**kwargs)


class BaseMsg:
    type: str

    def get_message(self):
        return self.__dict__()

    def get_value(self):
        return self.__dict__()["value"]

    def get_type(self):
        return self.type



import locale


def number_format(num, places=0):
    return locale.format_string("%.*f", (places, num), True)


def get_amount_uni(amount, reverse=False):
    exp = 18
    if reverse:
        exp = -18
    numb = amount * pow(10, exp)
    prepared_num = number_format(numb, 0)
    return prepared_num


def from_words(words):
    res = convert(words, 5, 8, False)
    return res


def convert(data, in_bits, out_bits, pad):
    value = 0
    bits = 0
    max_v = (1 << out_bits) - 1

    result = []
    i = 0
    while i < len(data):
        value = (value << in_bits) | data[i]
        bits += in_bits

        while bits >= out_bits:
            bits -= out_bits
            result.append((value >> bits) & max_v)
        i += 1

    if pad:
        if bits > 0:
            result.append((value << (out_bits - bits)) & max_v)
    else:
        if bits >= in_bits:
            return 'Excess padding'
        if (value << (out_bits - bits)) & max_v:
            return 'Non-zero padding'

    return result



import base64
import json
from hashlib import sha256

from ecdsa import SigningKey, SECP256k1
from ecdsa.util import sigencode_string_canonize

from .wallet import Wallet
from decimal_sdk.msgs.base import BaseMsg
from decimal_sdk.utils import beautify_json
from .tx_types import *

FEES = {
    COIN_SEND: 10,
    COIN_BUY: 10,
    COIN_CREATE: 100,
    COIN_SELL: 100,
    COIN_MULTISEND: 8,
    COIN_SELL_ALL: 100,
    COIN_REDEEM_CHECK: 30,
    VALIDATOR_CANDIDATE: 10000,
    VALIDATOR_DELEGATE: 200,
    VALIDATOR_SET_ONLINE: 100,
    VALIDATOR_SET_OFFLINE: 100,
    VALIDATOR_UNBOND: 200,
    VALIDATOR_CANDIDATE_EDIT: 10000,
    MULTISIG_CREATE_WALLET: 100,
    MULTISIG_CREATE_TX: 100,
    MULTISIG_SIGN_TX: 100,
    PROPOSAL_SUBMIT: 0,
    PROPOSAL_VOTE: 0,
    SWAP_INIT: 33000,
    SWAP_REDEEM: 0,
    COIN_UPDATE: 0,
    NFT_MINT: 0,
    NFT_BURN: 0,
    NFT_EDIT_METADATA: 0,
    NFT_TRANSFER: 0,
    NFT_DELEGATE: 0,
    NFT_UNBOND: 0,
    NFT_UPDATE_RESERVE: 0,
}


class Candidate:
    moniker: str
    identity: str
    website: str
    security_contact: str
    details: str

    def __init__(self, moniker: str, identity: str, website: str, security_contact: str, details: str):
        self.moniker = moniker
        self.identity = identity
        self.website = website
        self.security_contact = security_contact
        self.details = details

    def __dict__(self):
        return {"moniker": self.moniker, "identity": self.identity, "website": self.website,
                "security_contact": self.security_contact, "details": self.details}


class Coin:
    denom: str
    amount: str

    def __init__(self, denom: str, amount: str):
        self.denom = denom
        self.amount = amount

    def __dict__(self):
        return {'denom': self.denom, 'amount': str(self.amount)}


class Signature:
    pub_key: str
    signature: str

    def __init__(self, signature: str, pub_key: str):
        self.pub_key = pub_key
        self.signature = signature

    def __dict__(self):
        return {'pub_key': {'type': 'tendermint/PubKeySecp256k1', 'value': self.pub_key},
                'signature': self.signature}

    def get_signature(self):
        return self.__dict__()


class Fee:
    amount: [Coin]
    gas: str

    def __init__(self, amount: [Coin], gas: str):
        self.amount = amount
        self.gas = gas

    def __dict__(self):
        return {'gas': self.gas, 'amount': [coin.__dict__() for coin in self.amount]}


class SignMeta:
    account_number: str
    chain_id: str
    sequence: str

    def __init__(self, account_number: str = '0', chain_id: str = 'decimal-testnet-12-09-13-00', sequence: str = '1'):
        self.account_number = account_number
        self.chain_id = chain_id
        self.sequence = sequence


class StdSignMsg:
    # Tx part
    fee: Fee
    memo: str
    msgs: [BaseMsg]
    # Meta part
    account_number: str
    chain_id: str
    sequence: str

    signatures: [Signature] = []

    def __init__(self, tx, meta: SignMeta):
        self.fee = tx.fee
        self.msgs = tx.msgs
        self.memo = tx.memo

        self.account_number = meta.account_number
        self.chain_id = meta.chain_id
        self.sequence = meta.sequence

    def add_msg(self, msg: BaseMsg):
        self.msgs.append(msg)

    def sign(self, wallet: Wallet):
        private_key = wallet.get_private_key()
        pub_key = wallet.get_public_key()
        sig = self.__generate_signature(private_key)
        self.signatures = [Signature(signature=sig, pub_key=pub_key)]

    def __get_body_bytes(self):
        data = beautify_json(
            {'fee': self.fee.__dict__(), 'memo': self.memo, 'msgs': [msg.__dict__() for msg in self.msgs],
             'account_number': self.account_number, 'chain_id': self.chain_id, 'sequence': self.sequence})

        return json.dumps(data, separators=(',', ':'), ensure_ascii = False).encode('utf-8')

    def __generate_signature(self, private_key):
        data = self.__get_body_bytes()
        hash_ = sha256(data).digest()
        sk = SigningKey.from_string(private_key, curve=SECP256k1)
        signature = sk.sign_digest_deterministic(hash_, hashfunc=sha256, sigencode=sigencode_string_canonize)
        base_signature = base64.b64encode(signature)

        return base_signature.decode('utf-8')

    def __dict__(self):
        return {'fee': self.fee.__dict__(), 'memo': self.memo, 'msgs': [msg.__dict__() for msg in self.msgs],
                'account_number': self.account_number, 'chain_id': self.chain_id, 'sequence': self.sequence,
                'signatures': [sig.__dict__() for sig in self.signatures]}



import bech32


def verify_address(address, prefix="dx"):
    if not address:
        raise Exception("address is missing")

    try:
        decoded = bech32.bech32_decode(address)

        print(f"checked {address}, result is:")
        print(decoded[0] == prefix and decoded[1] != [])
        return decoded[0] == prefix and decoded[1] != []

    except:
        return False


def validate_data(data):
    if "from" in data:
        if data["from"] == "":
            raise Exception("Field \'from\' can not be empty")

        if not verify_address(data["from"]):
            raise Exception("Incorrect sender address format")

    if "sender" in data:
        if data["sender"] == "":
            raise Exception("Field \'sender\' can not be empty")

        if not verify_address(data["sender"]):
            raise Exception("Incorrect sender address format")

    if "to" in data:
        if data["to"] == "":
            raise Exception("Field \'to\' can not be empty")

        if not verify_address(data["to"]):
            raise Exception("Incorrect recipient address format")

    if "receiver" in data:
        if data["receiver"] == "":
            raise Exception("Field \'receiver\' can not be empty")

        if not verify_address(data["receiver"]):
            raise Exception("Incorrect receiver address format")

    if "validator_address" in data:
        if data["validator_address"] == "":
            raise Exception("Field \'validator_address\' can not be empty")

        if not verify_address(data["validator_address"], "dxvaloper"):
            raise Exception("Incorrect validator address format")

    if "delegator_address" in data:
        if data["delegator_address"] == "":
            raise Exception("Field \'delegator_address\' can not be empty")

        if not verify_address(data["delegator_address"]):
            raise Exception("Incorrect delegator address format")

    if "owners" in data:
        if not data["owners"]:
            raise Exception("Field \'owners\' can not be empty")

        for owner in data["owners"]:
            # if owner == "":
            #     raise Exception("Owner in field \'owners\' can not be empty")

            if not verify_address(owner):
                raise Exception(f"Incorrect owner address format: {owner}")

    if "sends" in data:
        if not data["sends"]:
            raise Exception("Field \'sends\' can not be empty")

        for send in data["sends"]:
            # if send == "":
            #     raise Exception("Recipient in field \'sends\' can not be empty")

            if not verify_address(send["receiver"]):
                raise Exception(f"Incorrect recipient address format: {send}")





import base64

from Cryptodome.Hash import SHA256, RIPEMD160
from mnemonic import Mnemonic
from bip32 import BIP32
import bech32

DERIVATION_PATH = "m/44'/60'/0'/0/0"
VALIDATOR_PREFIX = 'dxvaloper'


class Wallet:
    """
    Decimal wallet class
    """

    def __init__(self, mnemonic: str = None):
        mnemo = Mnemonic('english')
        if not mnemonic:
            mnemonic = mnemo.generate()

        if not mnemo.check(mnemonic):
            raise Exception('Invalid Mnemonic')
        self.__mnemonic = mnemonic

        # Generate and save seed
        seed = mnemo.to_seed(mnemonic)
        self._seed = seed
        # Derive public key
        self.__generate_keys()
        self.__generate_address()
        self.__generate_validator_address()

    def get_address(self) -> str:
        '''
        returns addres of the wallet
        '''
        return self._address

    def get_mnemonic(self) -> str:
        '''
        returns mnemonic of the wallet
        '''
        return self.__mnemonic

    def get_validator_address(self) -> str:
        '''
        returns validator of the wallet
        '''
        return self._validator_address

    def get_private_key(self) -> str:
        '''
        generates private key
        '''
        return self._private_key

    def get_public_key(self) -> str:
        '''
        generates public key
        '''
        return self._public_key

    def __generate_address(self):
        prepared_hash = self.__hash_public_key()
        address = bech32.bech32_encode('dx', bech32.convertbits(prepared_hash, 8, 5))
        self._address = address

    def __generate_validator_address(self):
        prepared_hash = self.__hash_public_key()
        address = bech32.bech32_encode(VALIDATOR_PREFIX, bech32.convertbits(prepared_hash, 8, 5))
        self._validator_address = address

    def __generate_keys(self):
        bip32 = BIP32.from_seed(self._seed)
        self._public_key = base64.b64encode(bip32.get_pubkey_from_path(DERIVATION_PATH)).decode('utf-8')
        self._public_key_binary = bip32.get_pubkey_from_path(DERIVATION_PATH)
        self._private_key = bip32.get_privkey_from_path(DERIVATION_PATH)

    def __hash_public_key(self):
        sha256_hash = SHA256.new(self._public_key_binary).digest()
        ripemd160_hash = RIPEMD160.new(sha256_hash).digest()
        return ripemd160_hash




import hashlib
import rlp
import ethereum.transactions as crypto


class Check:
    chain_id: str
    coin: str
    amount: int
    nonce: bytearray
    due_block: int
    lock: int
    v: int
    r: int
    s: int

    def get_sender(self):
        return self.recover_plain(self.check_hash(), self.r, self.s, self.v)

    def lock_pub_key_(self):
        sig = self.lock.to_bytes()
        if len(sig) < 65:
            sig = 65-len(sig) + int(sig)

        check_hash = self.hash_without_lock()
        pub = crypto.ecrecover_to_pub(check_hash, sig)

        if len(pub) == 0 or pub[0] != 4:
            raise Exception("Invalid public key")

        return pub

    def hash_without_lock(self):
        return self.rpl_hash(
            [
                self.chain_id,
                self.coin,
                self.amount,
                self.nonce,
                self.due_block
            ]
        )

    def check_hash(self):
        return self.rpl_hash(
            [
                self.chain_id,
                self.coin,
                self.amount,
                self.nonce,
                self.due_block,
                self.lock
            ]
        )

    def hash_full(self):
        return self.rpl_hash(
            [
                self.chain_id,
                self.coin,
                self.amount,
                self.nonce,
                self.due_block,
                self.lock,
                self.v,
                self.r,
                self.s
            ]
        )

    def sign(self, prvt_key):
        h = self.check_hash()
        sig = crypto.ecsign(h[:], prvt_key)
        self.set_signature(sig)

    def set_signature(self, sig: bytearray):
        self.r = sig[0:32]
        self.s = sig[32:64]
        self.v = int((sig[64] + 27).to_bytes())

    def recover_plain(self, sighash, r, s, v):
        pass

    @staticmethod
    def parse_check(buf):
        check: Check = rlp.decode(buf)
        if check.s == 0 or check.r == 0 or check.v == 0:
            raise Exception("Incorrect tx signature")

        return check

    @staticmethod
    def rpl_hash(data):
        return hashlib.sha3_256(data)

    def __str__(self):
        return f"Check sender: {self.get_sender()} nonce: {self.nonce}, " \
               f"dueBlock: {self.due_block}, value: {self.amount} {self.coin}"


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="decimal-python-sdk",
    version="0.1.2",
    author="DecimalTeam",
    description="",
    long_description=long_description,
    install_requires=['asn1crypto==1.4.0', 'base58==2.0.0', 'base58check==1.0.2', 'bech32==1.2.0', 'bip32==0.0.8', 'cached-property==1.5.2', 'certifi==2020.11.8', 'cffi==1.14.3', 'chardet==3.0.4', 'coincurve==13.0.0', 'cytoolz==0.11.0', 'ecdsa==0.16.1', 'eth-hash==0.3.1', 'eth-typing==2.2.2', 'eth-utils==1.10.0', 'ethereum==2.3.2', 'future==0.18.2', 'idna==2.10', 'mnemonic==0.19', 'mypy-extensions==0.4.3', 'pbkdf2==1.3', 'py-ecc==5.1.0', 'pyaes==1.6.1', 'pycparser==2.20', 'pycryptodome==3.9.9', 'pycryptodomex==3.9.9', 'pyethash==0.1.27', 'pysha3==1.0.2', 'PyYAML==5.4.1', 'repoze.lru==0.7', 'requests==2.25.0', 'rlp==1.2.0', 'rpl==1.8.1', 'scrypt==0.8.17', 'six==1.15.0', 'sslcrypto==2.0', 'toolz==0.11.1', 'urllib3==1.26.2'],
    long_description_content_type="text/markdown",
    url="https://bitbucket.org/decimalteam/decimal-python-sdk",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



COIN_BUY = 'coin/buy_coin'
COIN_CREATE = 'coin/create_coin'
COIN_UPDATE = 'coin/update_coin'
COIN_SELL = 'coin/sell_coin'
COIN_SEND = 'coin/send_coin'
COIN_MULTISEND = 'coin/multi_send_coin'
COIN_SELL_ALL = 'coin/sell_all_coin'
COIN_REDEEM_CHECK = 'coin/redeem_check'
COIN_ISSUE_CHECK = 'coin/issue_check'

VALIDATOR_CANDIDATE = 'validator/declare_candidate'
VALIDATOR_DELEGATE = 'validator/delegate'
VALIDATOR_SET_ONLINE = 'validator/set_online'
VALIDATOR_SET_OFFLINE = 'validator/set_offline'
VALIDATOR_UNBOND = 'validator/unbond'
VALIDATOR_CANDIDATE_EDIT = 'validator/edit_candidate'

MULTISIG_CREATE_WALLET = 'multisig/create_wallet'
MULTISIG_CREATE_TX = 'multisig/create_transaction'
MULTISIG_SIGN_TX = 'multisig/sign_transaction'

PROPOSAL_SUBMIT = 'cosmos-sdk/MsgSubmitProposal'
PROPOSAL_VOTE = 'cosmos-sdk/MsgVote'

NFT_MINT = 'nft/msg_mint'
NFT_BURN = 'nft/msg_burn'
NFT_EDIT_METADATA = 'nft/msg_edit_metadata'
NFT_TRANSFER = 'nft/msg_transfer'
NFT_DELEGATE = 'validator/delegate_nft'
NFT_UNBOND = 'validator/unbond_nft'
NFT_UPDATE_RESERVE = 'nft/update_reserve'

SWAP_INIT = 'swap/msg_initialize'
SWAP_REDEEM = 'swap/msg_redeem_v2'


from .api import DecimalAPI
from .wallet import Wallet
from .tx_types import *
from .transactions import (Transaction,
                           SendCoinTransaction, BuyCoinTransaction, CreateCoinTransaction, UpdateCoinTransaction, SellAllCoinsMsgTransaction,
                           DelegateTransaction, UnbondTransaction,
                           DeclareCandidateTransaction, EditCandidateTransaction,
                           DisableEnableValidatorTransaction,
                           MultysigCreateTransaction, MultysigCreateTXTransaction, MultysigSignTXTransaction, MultisendCoinTransaction,
                           SubmitProposalTransaction, VoteProposalTransaction,
                           SwapRedeemTransaction, SwapInitTransaction,
                           NftMintTransaction, NftBurnTransaction, NftEditMetadataTransaction, NftTransferTransaction, NftDelegateTransaction, NftUnboundTransaction, NftUpdateReserveTransaction,
                           )
